#include "pw/visual/renderer.hpp"

#include "pw/core/geometry.hpp"
#include "pw/core/size.hpp"
#include "pw/core/debug.hpp"

#include "glad/glad.h"

#include <algorithm>


namespace pw {

struct renderer::impl {

    GLuint _vao = 0;
    std::vector<GLuint> _vbos;
    GLint _mesh_elements = { 0 };

    impl() = default;

    ~impl()
    {
        release();
    }

    bool ready() const
    {
        return glIsVertexArray != nullptr && GL_TRUE == glIsVertexArray(_vao);
    }

    bool create(const geometry& m)
    {
        // reset if the renderer already in use
        if (ready()) {
            release();
        }

        glGenVertexArrays(1,&_vao);
        // TODO check if valid?
        glBindVertexArray(_vao);

        size_t arrays_needed = 0;

        // should bail out here
        if (!m.vertices().empty()) arrays_needed++;
        if (!m.indices().empty()) arrays_needed++;


        _mesh_elements = m.indices().size();

        // TODO: implement the other arrays

        _vbos.resize(arrays_needed);

        glGenBuffers(_vbos.size(), _vbos.data());

        // vertices
        glBindBuffer(GL_ARRAY_BUFFER, _vbos[0]);
        glBufferData(GL_ARRAY_BUFFER,
                     sizeof(m.vertices().front()) * m.vertices().size(),
                     m.vertices().data(),
                     GL_STATIC_DRAW);

        glEnableVertexAttribArray(0);
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, nullptr);

        // indices
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _vbos[1]);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER,
                     sizeof(m.indices().front()) * m.indices().size(), m.indices().data(),
                     GL_STATIC_DRAW);

        // stop binding
        glBindVertexArray(0);

        auto error = glGetError();
        if (error != GL_NO_ERROR) {
            debug::e() << "GL error: " << error;
        }

        return ready();
    }


    void release()
    {
        glDeleteVertexArrays(1,&_vao);
        for (auto vbo : _vbos)
            glDeleteBuffers(1,&vbo);

        _vbos.clear();
    }

    void draw()
    {
#if 1
        glEnable(GL_CULL_FACE);
        glCullFace(GL_BACK);
        glFrontFace(GL_CCW);
#endif

        glClearColor(0.0,1.0,0,1);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        if (glIsVertexArray(_vao))
        {

            glBindVertexArray(_vao);
            glDrawElements(GL_TRIANGLES, _mesh_elements, GL_UNSIGNED_INT, nullptr);
            glBindVertexArray(0);

            auto error = glGetError();
            if (error != GL_NO_ERROR)
            {
                debug::e() << "GL error: " << error;
            }
        }
    }

    //	GLint get_mode(vertex_array::)

};

//
//
//

renderer::renderer()
    : _impl(std::make_unique<renderer::impl>())
{
}

renderer::renderer(const geometry &m)
{
    renderer();
    _impl->create(m);
}

renderer::~renderer()
{
}

bool renderer::ready() const
{
    return _impl->ready();
}

bool renderer::create(const geometry &m)
{
    return _impl->create(m);
}

void renderer::release()
{
    _impl->release();
}

void renderer::draw()
{
    _impl->draw();
}


//class viewport {



//    void set(point p,size s);



//protected:

//    struct _impl;
//    std::unique_ptr<_impl>;

//};


}
